<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <title>KRCG API</title>
    <link rel="stylesheet" type="text/css" href="https://unpkg.com/swagger-ui-dist@3/swagger-ui.css">
    <script src="https://unpkg.com/swagger-ui-dist@3/swagger-ui-bundle.js"> </script>
    <script>
    window.onload = function() {
        window.ui = SwaggerUIBundle({
            url: "openapi.yaml",
            dom_id: '#swagger-ui',
        })
    }
    </script>
</head>
<body>
    <div id="swagger-ui"></div>
</body>
</html>
